local ADDON_NAME = ...;
local ADDON_VERSION = GetAddOnMetadata(ADDON_NAME, "Version");

local ADDON_PREFIX = "ezCollections";
local ITEM_HIDDEN = 15;
local ITEM_BACK = 16;
TRANSMOGRIFY_FONT_COLOR = "|cFFFF80FF";

local oGetInventoryItemID = GetInventoryItemID;

-- ---------
-- Ace Addon
-- ---------
local addon = LibStub("AceAddon-3.0"):NewAddon(ADDON_NAME, "AceEvent-3.0", "AceTimer-3.0");
local L = LibStub("AceLocale-3.0"):GetLocale(ADDON_NAME);

function addon:OnInitialize()
    BINDING_HEADER_EZCOLLECTIONS                    = L["Binding.Header"];
    BINDING_NAME_EZCOLLECTIONS_UNLOCK_SKIN          = L["Binding.UnlockSkin"];
    BINDING_NAME_EZCOLLECTIONS_MENU_ISENGARD        = L["Binding.Menu.Isengard"];
    BINDING_NAME_EZCOLLECTIONS_MENU_TRANSMOG        = L["Binding.Menu.Transmog"];
    BINDING_NAME_EZCOLLECTIONS_MENU_TRANSMOG_SETS   = L["Binding.Menu.Transmog.Sets"];
    BINDING_NAME_EZCOLLECTIONS_MENU_COLLECTIONS     = L["Binding.Menu.Collections"];
    BINDING_NAME_EZCOLLECTIONS_MENU_DAILY           = L["Binding.Menu.Daily"];

    self:RegisterEvent("CHAT_MSG_ADDON");
    self:RegisterEvent("PLAYER_LOGIN");
    self:RegisterEvent("INSPECT_TALENT_READY");
    self:RegisterEvent("UNIT_INVENTORY_CHANGED");
    self:RegisterEvent("BANKFRAME_OPENED");

    local PRESET_COLORS =
    {
        pink        = { Name = L["Color.Pink"],     r =         1, g =       0.5, b = 1, a = 1, code = TRANSMOGRIFY_FONT_COLOR },
        blue        = { Name = L["Color.Blue"],     r = 0x88/0xFF, g = 0xAA/0xFF, b = 1, a = 1, code = "|cFF88AAFF" },
        gold        = { Name = L["Color.Gold"],     r = 0xFF/0xFF, g = 0xD2/0xFF, b = 0, a = 1, code = "|cFFFFD200" },
        white       = { Name = L["Color.White"],    r =         1, g =         1, b = 1, a = 1, code = "|cFFFFFFFF" },
    };

    local defaultsConfig =
    {
        profile =
        {
            NewVersion =
            {
                HideRetiredPopup = false,
            },
            Alerts =
            {
                AddSkin =
                {
                    Enable = true,
                    Color = { Custom = false, r = PRESET_COLORS["pink"].r, g = PRESET_COLORS["pink"].g, b = PRESET_COLORS["pink"].b, a = PRESET_COLORS["pink"].a },
                    FullRowColor = false,
                },
            },
            TooltipTransmog =
            {
                Enable = true,
                IconEntry =
                {
                    Enable = true,
                    Size = 0,
                    Crop = true,
                },
                IconEnchant =
                {
                    Enable = true,
                    Size = 0,
                },
                Color = { Custom = false, r = PRESET_COLORS["pink"].r, g = PRESET_COLORS["pink"].g, b = PRESET_COLORS["pink"].b, a = PRESET_COLORS["pink"].a },
            },
            TooltipCollection =
            {
                OwnedItems = false,
                Skins = true,
                SkinUnlock = true,
                TakenQuests = false,
                RewardedQuests = false,
                Color = { Custom = false, r = PRESET_COLORS["blue"].r, g = PRESET_COLORS["blue"].g, b = PRESET_COLORS["blue"].b, a = PRESET_COLORS["blue"].a },
                Separator = true,
            },
            RestoreItemIcons =
            {
                Equipment = true,
                Inspect = true,
            },
            RestoreItemSets =
            {
                Equipment = true,
            },
        },
    };
    local defaultsCache =
    {
        realm =
        {
            Version = 0,
            All = { },
            Slot =
            {
                ["HEAD"]            = { },
                ["SHOULDER"]        = { },
                ["BACK"]            = { },
                ["CHEST"]           = { },
                ["TABARD"]          = { },
                ["SHIRT"]           = { },
                ["WRIST"]           = { },
                ["HANDS"]           = { },
                ["WAIST"]           = { },
                ["LEGS"]            = { },
                ["FEET"]            = { },
                ["WAND"]            = { },
                ["1H_AXE"]          = { },
                ["1H_SWORD"]        = { },
                ["1H_MACE"]         = { },
                ["DAGGER"]          = { },
                ["FIST"]            = { },
                ["SHIELD"]          = { },
                ["HOLDABLE"]        = { },
                ["2H_AXE"]          = { },
                ["2H_SWORD"]        = { },
                ["2H_MACE"]         = { },
                ["STAFF"]           = { },
                ["POLEARM"]         = { },
                ["BOW"]             = { },
                ["GUN"]             = { },
                ["CROSSBOW"]        = { },
                ["THROWN"]          = { },
                ["FISHING_POLE"]    = { },
                ["MISC"]            = { },
                ["ENCHANT"]         = { },
            },
        },
    };

    ezCollections.Config = LibStub("AceDB-3.0"):New(ADDON_NAME.."Config", defaultsConfig, true).profile;
    local cache = LibStub("AceDB-3.0"):New(ADDON_NAME.."Cache", defaultsCache, true);
    ezCollections.Cache = cache.realm;

    ezCollections.ClearCache = function(self)
        cache:ResetDB(nil);
        self.Cache = cache.realm;
    end;

    local keybindHandler =
    {
        Get = function(self, info)
            return table.concat({ GetBindingKey(info.arg) }, ", ");
        end,
        Set = function(self, info, value)
            if value == "" then value = nil; end

            if value and GetBindingAction(value) ~= "" and GetBindingAction(value) ~= info.arg then
                self:Error(L["Binding.Error.AlreadyBound"]);
                return;
            end

            if value then
                if not SetBinding(value, info.arg) then
                    self:Error(L["Binding.Error.BindingFailed"]);
                    return;
                end
            else
                local keys = { GetBindingKey(info.arg) };
                for _, key in pairs(keys) do
                    SetBinding(key);
                end
            end

            if GetCurrentBindingSet()==1 or GetCurrentBindingSet()==2 then
                SaveBindings(GetCurrentBindingSet());
            end
        end,
        Error = function(self, message)
            StaticPopupDialogs["EZCOLLECTIONS_KEYBINDING_ERROR"].text = message;
            StaticPopup_Show("EZCOLLECTIONS_KEYBINDING_ERROR");
        end,
    };
    LibStub("AceConfig-3.0"):RegisterOptionsTable(ADDON_NAME,
    {
        name = L["Addon.Color"],
        handler = addon,
        type = "group",
        args =
        {
            newVersion =
            {
                type = "group",
                name = function() return ezCollections.NewVersion == nil and "" or ezCollections.NewVersion.Disabled and L["Config.NewVersion.Disabled"] or ezCollections.NewVersion.Outdated and L["Config.NewVersion.Outdated"] or L["Config.NewVersion.Compatible"] end,
                inline = true,
                order = 0,
                hidden = function() return ezCollections.NewVersion == nil; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.NewVersion.Desc"],
                        order = 0,
                        hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                    },
                    url =
                    {
                        type = "input",
                        name = L["Config.NewVersion.URL"],
                        order = 1,
                        width = "full",
                        get = function(info) return ezCollections.NewVersion.URL; end;
                        set = function(info, value) end;
                        hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                    },
                    clientVersion =
                    {
                        type = "description",
                        name = function() return format(L["Config.NewVersion.ClientVersion"], ADDON_VERSION); end;
                        order = 2,
                        hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                    },
                    serverVersion =
                    {
                        type = "description",
                        name = function() return format(L["Config.NewVersion.ServerVersion"], ezCollections.NewVersion.Version); end;
                        order = 3,
                        hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                    },
                    hideRetiredPopup =
                    {
                        type = "toggle",
                        name = L["Config.NewVersion.HideRetiredPopup"],
                        desc = L["Config.NewVersion.HideRetiredPopup.Desc"],
                        width = "full",
                        order = 100,
                        get = function(info) return ezCollections.Config.NewVersion.HideRetiredPopup; end,
                        set = function(info, value) ezCollections.Config.NewVersion.HideRetiredPopup = value; end,
                        hidden = function() return ezCollections.NewVersion and not ezCollections.NewVersion.Disabled; end,
                    },
                },
            },
            alerts =
            {
                type = "group",
                name = L["Config.Alerts"],
                inline = true,
                order = 100,
                hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.Alerts.Desc"],
                        order = 0,
                    },
                    enable =
                    {
                        type = "toggle",
                        name = L["Config.Alerts.AddSkin.Enable"],
                        desc = L["Config.Alerts.AddSkin.Enable.Desc"],
                        width = "full",
                        order = 100,
                        get = function(info) return ezCollections.Config.Alerts.AddSkin.Enable; end,
                        set = function(info, value) ezCollections.Config.Alerts.AddSkin.Enable = value; end,
                    },
                    color =
                    {
                        type = "select",
                        name = L["Config.Alerts.AddSkin.Color"],
                        desc = L["Config.Alerts.AddSkin.Color.Desc"],
                        values = function()
                            local values = { custom = L["Color.Custom"] };
                            for k,v in pairs(PRESET_COLORS) do
                                values[k] = v.Name;
                            end
                            return values;
                        end,
                        order = 101,
                        get = function(info)
                            local color = ezCollections.Config.Alerts.AddSkin.Color;
                            if not color.Custom then
                                for k,v in pairs(PRESET_COLORS) do
                                    if v.r == color.r and
                                       v.g == color.g and
                                       v.b == color.b and
                                       v.a == color.a then
                                        return k;
                                    end
                                end
                            end
                            return "custom";
                        end,
                        set = function(info, value)
                            local color = ezCollections.Config.Alerts.AddSkin.Color;
                            if value == "custom" then
                                color.Custom = true;
                            else
                                color.Custom = false;
                                color.r = PRESET_COLORS[value].r;
                                color.g = PRESET_COLORS[value].g;
                                color.b = PRESET_COLORS[value].b;
                                color.a = PRESET_COLORS[value].a;
                            end
                        end,
                        disabled = function() return not ezCollections.Config.Alerts.AddSkin.Enable; end,
                    },
                    customColor =
                    {
                        type = "color",
                        name = L["Config.Alerts.AddSkin.CustomColor"],
                        desc = L["Config.Alerts.AddSkin.CustomColor.Desc"],
                        width = "half",
                        order = 102,
                        get = function(info)             local color = ezCollections.Config.Alerts.AddSkin.Color; return color.r,     color.g,     color.b,     color.a;     end,
                        set = function(info, r, g, b, a) local color = ezCollections.Config.Alerts.AddSkin.Color;        color.r = r; color.g = g; color.b = b; color.a = a; end,
                        disabled = function() return not ezCollections.Config.Alerts.AddSkin.Enable or not ezCollections.Config.Alerts.AddSkin.Color.Custom; end,
                    },
                    fullRowColor =
                    {
                        type = "toggle",
                        name = L["Config.Alerts.AddSkin.FullRowColor"],
                        desc = L["Config.Alerts.AddSkin.FullRowColor.Desc"],
                        width = "half",
                        order = 103,
                        get = function(info) return ezCollections.Config.Alerts.AddSkin.FullRowColor; end,
                        set = function(info, value) ezCollections.Config.Alerts.AddSkin.FullRowColor = value; end,
                        disabled = function() return not ezCollections.Config.Alerts.AddSkin.Enable; end,
                    },
                },
            },
            tooltipTransmog =
            {
                type = "group",
                name = L["Config.TooltipTransmog"],
                inline = true,
                order = 200,
                hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.TooltipTransmog.Desc"],
                        order = 0,
                    },
                    enable =
                    {
                        type = "toggle",
                        name = L["Config.TooltipTransmog.Enable"],
                        desc = L["Config.TooltipTransmog.Enable.Desc"],
                        width = "full",
                        order = 1,
                        get = function(info) return ezCollections.Config.TooltipTransmog.Enable; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.Enable = value; end,
                    },
                    iconEntry =
                    {
                        type = "toggle",
                        name = L["Config.TooltipTransmog.IconEntry"],
                        desc = L["Config.TooltipTransmog.IconEntry.Desc"],
                        width = "normal",
                        order = 100,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEntry.Enable; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEntry.Enable = value; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable; end,
                    },
                    iconEntrySize =
                    {
                        type = "range",
                        name = L["Config.TooltipTransmog.IconEntry.Size"],
                        desc = L["Config.TooltipTransmog.IconEntry.Size.Desc"],
                        min = 0,
                        max = 64,
                        softMin = 8,
                        bigStep = 1,
                        width = "half",
                        order = 101,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEntry.Size; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEntry.Size = value; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable or not ezCollections.Config.TooltipTransmog.IconEntry.Enable or ezCollections.Config.TooltipTransmog.IconEntry.Size == 0; end,
                    },
                    iconEntrySizeAuto =
                    {
                        type = "toggle",
                        name = L["Config.TooltipTransmog.IconEntry.Size.Auto"],
                        desc = L["Config.TooltipTransmog.IconEntry.Size.Auto.Desc"],
                        width = "half",
                        order = 102,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEntry.Size == 0; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEntry.Size = value and 0 or 16; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable or not ezCollections.Config.TooltipTransmog.IconEntry.Enable; end,
                    },
                    iconEntryPadding1 =
                    {
                        type = "description",
                        name = "",
                        width = "normal",
                        order = 103,
                    },
                    iconEntryPadding2 =
                    {
                        type = "description",
                        name = "",
                        width = "half",
                        order = 104,
                    },
                    iconEntryCrop =
                    {
                        type = "toggle",
                        name = L["Config.TooltipTransmog.IconEntry.Crop"],
                        desc = L["Config.TooltipTransmog.IconEntry.Crop.Desc"],
                        width = "half",
                        order = 105,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEntry.Crop; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEntry.Crop = value; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable or not ezCollections.Config.TooltipTransmog.IconEntry.Enable; end,
                    },
                    iconEnchant =
                    {
                        type = "toggle",
                        name = L["Config.TooltipTransmog.IconEnchant"],
                        desc = L["Config.TooltipTransmog.IconEnchant.Desc"],
                        width = "normal",
                        order = 200,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEnchant.Enable; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEnchant.Enable = value; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable; end,
                    },
                    iconEnchantSize =
                    {
                        type = "range",
                        name = L["Config.TooltipTransmog.IconEnchant.Size"],
                        desc = L["Config.TooltipTransmog.IconEnchant.Size.Desc"],
                        min = 0,
                        max = 64,
                        softMin = 8,
                        bigStep = 1,
                        width = "half",
                        order = 201,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEnchant.Size; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEnchant.Size = value; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable or not ezCollections.Config.TooltipTransmog.IconEnchant.Enable or ezCollections.Config.TooltipTransmog.IconEnchant.Size == 0; end,
                    },
                    iconEnchantSizeAuto =
                    {
                        type = "toggle",
                        name = L["Config.TooltipTransmog.IconEnchant.Size.Auto"],
                        desc = L["Config.TooltipTransmog.IconEnchant.Size.Auto.Desc"],
                        width = "half",
                        order = 202,
                        get = function(info) return ezCollections.Config.TooltipTransmog.IconEnchant.Size == 0; end,
                        set = function(info, value) ezCollections.Config.TooltipTransmog.IconEnchant.Size = value and 0 or 16; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable or not ezCollections.Config.TooltipTransmog.IconEnchant.Enable; end,
                    },
                    color =
                    {
                        type = "select",
                        name = L["Config.TooltipTransmog.Color"],
                        desc = L["Config.TooltipTransmog.Color.Desc"],
                        values = function()
                            local values = { custom = L["Color.Custom"] };
                            for k,v in pairs(PRESET_COLORS) do
                                values[k] = v.Name;
                            end
                            return values;
                        end,
                        order = 300,
                        get = function(info)
                            local color = ezCollections.Config.TooltipTransmog.Color;
                            if not color.Custom then
                                for k,v in pairs(PRESET_COLORS) do
                                    if v.r == color.r and
                                       v.g == color.g and
                                       v.b == color.b and
                                       v.a == color.a then
                                        return k;
                                    end
                                end
                            end
                            return "custom";
                        end,
                        set = function(info, value)
                            local color = ezCollections.Config.TooltipTransmog.Color;
                            if value == "custom" then
                                color.Custom = true;
                            else
                                color.Custom = false;
                                color.r = PRESET_COLORS[value].r;
                                color.g = PRESET_COLORS[value].g;
                                color.b = PRESET_COLORS[value].b;
                                color.a = PRESET_COLORS[value].a;
                            end
                        end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable; end,
                    },
                    customColor =
                    {
                        type = "color",
                        name = L["Config.TooltipTransmog.CustomColor"],
                        desc = L["Config.TooltipTransmog.CustomColor.Desc"],
                        order = 301,
                        get = function(info)             local color = ezCollections.Config.TooltipTransmog.Color; return color.r,     color.g,     color.b,     color.a;     end,
                        set = function(info, r, g, b, a) local color = ezCollections.Config.TooltipTransmog.Color;        color.r = r; color.g = g; color.b = b; color.a = a; end,
                        disabled = function() return not ezCollections.Config.TooltipTransmog.Enable or not ezCollections.Config.TooltipTransmog.Color.Custom; end,
                    },
                },
            },
            tooltipCollection =
            {
                type = "group",
                name = L["Config.TooltipCollection"],
                inline = true,
                order = 300,
                hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.TooltipCollection.Desc"],
                        order = 0,
                    },
                    ownedItems =
                    {
                        type = "toggle",
                        name = L["Config.TooltipCollection.OwnedItems"],
                        desc = L["Config.TooltipCollection.OwnedItems.Desc"],
                        width = "full",
                        order = 100,
                        get = function(info) return ezCollections.Config.TooltipCollection.OwnedItems; end,
                        set = function(info, value) ezCollections.Config.TooltipCollection.OwnedItems = value; end,
                        hidden = function() return not ezCollections.Collections.OwnedItems.Enabled; end,
                    },
                    skins =
                    {
                        type = "toggle",
                        name = L["Config.TooltipCollection.Skins"],
                        desc = L["Config.TooltipCollection.Skins.Desc"],
                        width = "full",
                        order = 200,
                        get = function(info) return ezCollections.Config.TooltipCollection.Skins; end,
                        set = function(info, value) ezCollections.Config.TooltipCollection.Skins = value; end,
                        hidden = function() return not ezCollections.Collections.Skins.Enabled; end,
                    },
                    skinUnlock =
                    {
                        type = "toggle",
                        name = L["Config.TooltipCollection.SkinUnlock"],
                        desc = L["Config.TooltipCollection.SkinUnlock.Desc"],
                        order = 201,
                        get = function(info) return ezCollections.Config.TooltipCollection.SkinUnlock; end,
                        set = function(info, value) ezCollections.Config.TooltipCollection.SkinUnlock = value; end,
                        hidden = function() return not ezCollections.Collections.Skins.Enabled; end,
                    },
                    skinUnlockBinding =
                    {
                        type = "keybinding",
                        name = L["Config.TooltipCollection.SkinUnlock.Binding"],
                        desc = L["Config.TooltipCollection.SkinUnlock.Binding.Desc"],
                        order = 202,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_UNLOCK_SKIN",
                        get = "Get",
                        set = "Set",
                        hidden = function() return not ezCollections.Collections.Skins.Enabled; end,
                    },
                    takenQuests =
                    {
                        type = "toggle",
                        name = L["Config.TooltipCollection.TakenQuests"],
                        desc = L["Config.TooltipCollection.TakenQuests.Desc"],
                        width = "full",
                        order = 300,
                        get = function(info) return ezCollections.Config.TooltipCollection.TakenQuests; end,
                        set = function(info, value) ezCollections.Config.TooltipCollection.TakenQuests = value; end,
                        hidden = function() return not ezCollections.Collections.TakenQuests.Enabled; end,
                    },
                    rewardedQuests =
                    {
                        type = "toggle",
                        name = L["Config.TooltipCollection.RewardedQuests"],
                        desc = L["Config.TooltipCollection.RewardedQuests.Desc"],
                        width = "full",
                        order = 400,
                        get = function(info) return ezCollections.Config.TooltipCollection.RewardedQuests; end,
                        set = function(info, value) ezCollections.Config.TooltipCollection.RewardedQuests = value; end,
                        hidden = function() return not ezCollections.Collections.RewardedQuests.Enabled; end,
                    },
                    color =
                    {
                        type = "select",
                        name = L["Config.TooltipCollection.Color"],
                        desc = L["Config.TooltipCollection.Color.Desc"],
                        values = function()
                            local values = { custom = L["Color.Custom"] };
                            for k,v in pairs(PRESET_COLORS) do
                                values[k] = v.Name;
                            end
                            return values;
                        end,
                        order = 1000,
                        get = function(info)
                            local color = ezCollections.Config.TooltipCollection.Color;
                            if not color.Custom then
                                for k,v in pairs(PRESET_COLORS) do
                                    if v.r == color.r and
                                       v.g == color.g and
                                       v.b == color.b and
                                       v.a == color.a then
                                        return k;
                                    end
                                end
                            end
                            return "custom";
                        end,
                        set = function(info, value)
                            local color = ezCollections.Config.TooltipCollection.Color;
                            if value == "custom" then
                                color.Custom = true;
                            else
                                color.Custom = false;
                                color.r = PRESET_COLORS[value].r;
                                color.g = PRESET_COLORS[value].g;
                                color.b = PRESET_COLORS[value].b;
                                color.a = PRESET_COLORS[value].a;
                            end
                        end,
                    },
                    customColor =
                    {
                        type = "color",
                        name = L["Config.TooltipCollection.CustomColor"],
                        desc = L["Config.TooltipCollection.CustomColor.Desc"],
                        width = "half",
                        order = 1001,
                        get = function(info)             local color = ezCollections.Config.TooltipCollection.Color; return color.r,     color.g,     color.b,     color.a;     end,
                        set = function(info, r, g, b, a) local color = ezCollections.Config.TooltipCollection.Color;        color.r = r; color.g = g; color.b = b; color.a = a; end,
                        disabled = function() return not ezCollections.Config.TooltipCollection.Color.Custom; end,
                    },
                    separator =
                    {
                        type = "toggle",
                        name = L["Config.TooltipCollection.Separator"],
                        desc = L["Config.TooltipCollection.Separator.Desc"],
                        width = "half",
                        order = 1002,
                        get = function(info) return ezCollections.Config.TooltipCollection.Separator; end,
                        set = function(info, value) ezCollections.Config.TooltipCollection.Separator = value; end,
                    },
                },
            },
            restoreItemIcons =
            {
                type = "group",
                name = L["Config.RestoreItemIcons"],
                inline = true,
                order = 400,
                hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.RestoreItemIcons.Desc"],
                        order = 0,
                    },
                    equipment =
                    {
                        type = "toggle",
                        name = L["Config.RestoreItemIcons.Equipment"],
                        desc = L["Config.RestoreItemIcons.Equipment.Desc"],
                        width = "full",
                        order = 100,
                        get = function(info) return ezCollections.Config.RestoreItemIcons.Equipment; end,
                        set = function(info, value) ezCollections.Config.RestoreItemIcons.Equipment = value; end,
                    },
                    inspect =
                    {
                        type = "toggle",
                        name = L["Config.RestoreItemIcons.Inspect"],
                        desc = L["Config.RestoreItemIcons.Inspect.Desc"],
                        width = "full",
                        order = 200,
                        get = function(info) return ezCollections.Config.RestoreItemIcons.Inspect; end,
                        set = function(info, value) ezCollections.Config.RestoreItemIcons.Inspect = value; end,
                    },
                },
            },
            restoreItemSets =
            {
                type = "group",
                name = L["Config.RestoreItemSets"],
                inline = true,
                order = 401,
                hidden = function() return ezCollections.NewVersion and ezCollections.NewVersion.Disabled; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.RestoreItemSets.Desc"],
                        order = 0,
                    },
                    equipment =
                    {
                        type = "toggle",
                        name = L["Config.RestoreItemSets.Equipment"],
                        desc = L["Config.RestoreItemSets.Equipment.Desc"],
                        width = "full",
                        order = 100,
                        get = function(info) return ezCollections.Config.RestoreItemSets.Equipment; end,
                        set = function(info, value) ezCollections.Config.RestoreItemSets.Equipment = value; end,
                    },
                },
            },
            bindings =
            {
                type = "group",
                name = L["Config.Bindings"],
                inline = true,
                order = 900,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.Bindings.Desc"],
                        order = 0,
                    },
                    skinUnlockDesc =
                    {
                        type = "description",
                        name = L["Binding.UnlockSkin"],
                        width = "normal",
                        order = 100,
                    },
                    skinUnlock =
                    {
                        type = "keybinding",
                        name = "",
                        order = 101,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_UNLOCK_SKIN",
                        get = "Get",
                        set = "Set",
                    },
                    menuIsengardDesc =
                    {
                        type = "description",
                        name = L["Binding.Menu.Isengard"],
                        width = "normal",
                        order = 200,
                    },
                    menuIsengard =
                    {
                        type = "keybinding",
                        name = "",
                        order = 201,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_MENU_ISENGARD",
                        get = "Get",
                        set = "Set",
                    },
                    menuTransmogDesc =
                    {
                        type = "description",
                        name = L["Binding.Menu.Transmog"],
                        width = "normal",
                        order = 300,
                    },
                    menuTransmog =
                    {
                        type = "keybinding",
                        name = "",
                        order = 301,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_MENU_TRANSMOG",
                        get = "Get",
                        set = "Set",
                    },
                    menuTransmogSetsDesc =
                    {
                        type = "description",
                        name = L["Binding.Menu.Transmog.Sets"],
                        width = "normal",
                        order = 350,
                    },
                    menuTransmogSets =
                    {
                        type = "keybinding",
                        name = "",
                        order = 351,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_MENU_TRANSMOG_SETS",
                        get = "Get",
                        set = "Set",
                    },
                    menuCollectionsDesc =
                    {
                        type = "description",
                        name = L["Binding.Menu.Collections"],
                        width = "normal",
                        order = 400,
                    },
                    menuCollections =
                    {
                        type = "keybinding",
                        name = "",
                        order = 401,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_MENU_COLLECTIONS",
                        get = "Get",
                        set = "Set",
                    },
                    menuDailyDesc =
                    {
                        type = "description",
                        name = L["Binding.Menu.Daily"],
                        width = "normal",
                        order = 500,
                    },
                    menuDaily =
                    {
                        type = "keybinding",
                        name = "",
                        order = 501,
                        handler = keybindHandler,
                        arg = "EZCOLLECTIONS_MENU_DAILY",
                        get = "Get",
                        set = "Set",
                    },
                },
            },
            cache =
            {
                type = "group",
                name = L["Config.Cache"],
                inline = true,
                order = 1000,
                hidden = function() return true; end,
                args =
                {
                    desc =
                    {
                        type = "description",
                        name = L["Config.Cache.Desc"],
                        order = 0,
                    },
                    clear =
                    {
                        type = "execute",
                        name = L["Config.Cache.Clear"],
                        desc = L["Config.Cache.Clear.Desc"],
                        width = "full",
                        order = 1,
                        func = function() ezCollections:ClearCache(); end,
                    },
                },
            },
        },
    });
    LibStub("AceConfigDialog-3.0"):AddToBlizOptions(ADDON_NAME, nil, nil);

    StaticPopupDialogs["EZCOLLECTIONS_NEW_VERSION"] =
    {
        text = "",
        button1 = OKAY,
        hasEditBox = 1,
        OnShow = function(self)
            self.text:SetText(format(ezCollections.NewVersion.Disabled and L["Popup.NewVersion.Disabled"] or ezCollections.NewVersion.Outdated and L["Popup.NewVersion.Outdated"] or L["Popup.NewVersion.Compatible"], ezCollections.NewVersion.Version));
            if ezCollections.NewVersion.Disabled then
                self.editBox:Hide();
            else
                self.editBox:SetText(ezCollections.NewVersion.URL);
                self.editBox:SetFocus();
                self.editBox:HighlightText();
            end
        end,
        EditBoxOnEnterPressed = function(self, data)
            self:GetParent():Hide();
        end,
        EditBoxOnEscapePressed = function(self)
            self:GetParent():Hide();
        end,
        timeout = 0,
        whileDead = 1,
        hideOnEscape = 1,
        hideOnEnter = 1,
        showAlert = 1,
    };
    StaticPopupDialogs["EZCOLLECTIONS_UNLOCK_SKIN"] =
    {
        text = "",
        button1 = YES,
        button2 = NO,
        OnShow = function(self)
            self.text:SetText(format(L["Popup.UnlockSkin"], StaticPopupDialogs["EZCOLLECTIONS_UNLOCK_SKIN"].itemLink));
        end,
        OnAccept = function(self)
            ezCollections:SendAddonMessage( "UNLOCKSKIN:"..StaticPopupDialogs["EZCOLLECTIONS_UNLOCK_SKIN"].commandData);
        end,
        timeout = 0,
        whileDead = 1,
        hideOnEscape = 1,
        itemLink = nil,
        commandData = nil,
    };
    StaticPopupDialogs["EZCOLLECTIONS_KEYBINDING_ERROR"] =
    {
        text = "",
        button1 = OKAY,
        timeout = 0,
        whileDead = 1,
        hideOnEscape = 1,
    };
end

-- ----------------
-- Helper functions
-- ----------------
local function RGBPercToHex(r, g, b)
    r = r <= 1 and r >= 0 and r or 0
    g = g <= 1 and g >= 0 and g or 0
    b = b <= 1 and b >= 0 and b or 0
    return string.format("%02X%02X%02X", r*255, g*255, b*255)
end
local function IsSameColor(color, r, g, b)
    return abs(color[1] - r) <= 0.01
       and abs(color[2] - g) <= 0.01
       and abs(color[3] - b) <= 0.01;
end
local function FormatToPattern(format)
    return (format:gsub("%(", "%%("):gsub("%)", "%%)"):gsub("%%s", "(.+)"):gsub("%%d", "(%%d+)"):gsub("|4(.+):%1.+:%1.+;", "%1.-"));
end
local function starts_with(str, start)
   return str:sub(1, #start) == start
end
local function ends_with(str, ending)
   return ending == "" or str:sub(-#ending) == ending
end
local function match(str, prefix, callback)
    if starts_with(str, prefix) then
        callback(str:sub(#prefix + 1));
        return true;
    end
    return false;
end
local function GetItemID(item)
    if type(item) == "number" then
        return item;
    else
        local _, link = GetItemInfo(item);
        local _, id, _ = strsplit(":", link);
        return tonumber(id);
    end
end

-- -----------------
-- Main core and API
-- -----------------
ezCollections =
{
    Name = ADDON_NAME,
    Version = ADDON_VERSION,
    AceAddon = addon,
    Config = nil,
    Cache = nil,
    ClearCache = nil,
    Allowed = false,

    -- Communications
    NewVersion = nil,
    SendAddonMessage = function(self, msg)
        SendAddonMessage(ADDON_PREFIX, msg, "WHISPER", UnitName("player"));
    end,
    SendAddonCommand = function(self, msg)
        SendAddonMessage(msg.." ", "", "WHISPER", UnitName("player"));
    end,

    -- Collections
    Collections =
    {
        OwnedItems = { },
        Skins = { },
        TakenQuests = { },
        RewardedQuests = { },
    },
    IsSkinSource     = function(self, item)  local db = self.Cache.All;                  if                   not db.Loaded then return nil; end return db[GetItemID(item)] or false; end,
    HasOwnedItem     = function(self, item)  local db = self.Collections.OwnedItems;     if not db.Enabled or not db.Loaded then return nil; end return db[GetItemID(item)] or false; end,
    HasSkin          = function(self, item)  local db = self.Collections.Skins;          if not db.Enabled or not db.Loaded then return nil; end return db[GetItemID(item)] or false; end,
    HasTakenQuest    = function(self, quest) local db = self.Collections.TakenQuests;    if not db.Enabled or not db.Loaded then return nil; end return db[quest] or false; end,
    HasRewardedQuest = function(self, quest) local db = self.Collections.RewardedQuests; if not db.Enabled or not db.Loaded then return nil; end return db[quest] or false; end,

    -- Alerts
    Alerts =
    {
        AddSkin = function(item)
            local config = ezCollections.Config.Alerts.AddSkin;
            if not config.Enable then return; end
            local text = ERR_LEARN_TRANSMOG_S;
            local color = config.Color;
            local colorHex = "|cFF"..RGBPercToHex(color.r, color.g, color.b);
            if config.FullRowColor then
                text = colorHex..text.."|r";
            else
                text = "|cFFFFFFFF"..format(text, "|r"..colorHex.."%s|r|cFFFFFFFF").."|r";
            end
            local name = GetItemInfo(item);
            if name then
                SendSystemMessage(format(text, format("|Hitem:%d:0:0:0:0:0:0:0:0|h[%s]|h", item, name or "")));
            else
                ezCollections.dataRequestTooltip:SetHyperlink("item:"..item);
                ezCollections.dataRequestTooltip:Hide();
                local handler = { };
                handler.func = function(arg)
                    local text, item = unpack(arg);
                    local name = GetItemInfo(item);
                    if name then
                        SendSystemMessage(format(text, format("|Hitem:%d:0:0:0:0:0:0:0:0|h[%s]|h", item, name or "")));
                    else
                        ezCollections.AceAddon:ScheduleTimer(handler.func, 1, arg);
                    end
                end;
                handler.func({ text, item });
            end
        end,
    },

    -- Bindings
    itemUnderCursor = { ID = nil, Bag = nil, Slot = nil },
    UnlockSkinHintCommand = "",
    UnlockSkinUnderCursor = function(self)
        if self.Allowed and self.itemUnderCursor.ID and self.itemUnderCursor.Bag then
            local _, link, _, _, _, _, _, _, _, texture = GetItemInfo(self.itemUnderCursor.ID);
            StaticPopupDialogs["EZCOLLECTIONS_UNLOCK_SKIN"].itemLink = "|T"..texture..":30:30:0:-8|t "..link;
            if self.itemUnderCursor.Slot then
                StaticPopupDialogs["EZCOLLECTIONS_UNLOCK_SKIN"].commandData = self.itemUnderCursor.ID.." "..self.itemUnderCursor.Bag.." "..(self.itemUnderCursor.Slot - 1);
            else
                StaticPopupDialogs["EZCOLLECTIONS_UNLOCK_SKIN"].commandData = self.itemUnderCursor.ID.." "..(self.itemUnderCursor.Bag - 1);
            end
            StaticPopup_Show("EZCOLLECTIONS_UNLOCK_SKIN");
        end
    end,
    MenuIsengard     = function(self) self:SendAddonCommand(".isengard"); end,
    MenuTransmog     = function(self) self:SendAddonCommand(".isengard transmog"); end,
    MenuTransmogSets = function(self) self:SendAddonCommand(".isengard transmog set"); end,
    MenuCollections  = function(self) self:SendAddonCommand(".isengard transmog collection"); end,
    MenuDaily        = function(self) self:SendAddonCommand(".isengard daily"); end,

    -- Item Transmog
    ItemTransmogCache = { },
    GetItemTransmogCache = function(self, unit, bag, slot)
        unit = GetUnitName(unit or "player") or unit;
        self.ItemTransmogCache[unit] = self.ItemTransmogCache[unit] or { };
        self.ItemTransmogCache[unit][bag] = self.ItemTransmogCache[unit][bag] or { };
        if slot ~= nil then
            self.ItemTransmogCache[unit][bag][slot] = self.ItemTransmogCache[unit][bag][slot] or { };
            return self.ItemTransmogCache[unit][bag][slot];
        else
            return self.ItemTransmogCache[unit][bag];
        end
    end,
    ClearItemTransmogCache = function(self, unit)
        if UnitIsUnit(unit, "player") then return; end
        unit = GetUnitName(unit or "player") or unit;
        self.ItemTransmogCache[unit] = { };
    end,
    GetItemTransmog = function(self, unit, bag, slot)
        local id, request;
        if slot ~= nil then
            id = GetContainerItemID(bag, slot);
            request = bag.." "..(slot - 1);
        else
            id = oGetInventoryItemID(unit, bag);
            request = tostring(bag - 1);
        end
        if not id then return; end

        local cache = self:GetItemTransmogCache(unit, bag, slot);
        if cache.Loaded and (cache.ID == id or cache.FakeEntry == id and GetUnitName(unit) ~= GetUnitName("player")) then -- Upon inspect, GetInventoryItemID returns visible item IDs, i.e. fake transmogrified entries
            return cache.FakeEntry, cache.FakeEnchant;
        elseif (not cache.Loading or cache.ID ~= id) and GetUnitName(unit) == GetUnitName("player") then
            cache.ID = id;
            cache.Loaded = false;
            cache.Loading = true;
            self:SendAddonMessage("GETTRANSMOG:"..request);
        end
    end,
    setEmptyItemTransmogCache = false,
    SetEmptyItemTransmogCache = function(self)
        if self.setEmptyItemTransmogCache then return; end
        self.setEmptyItemTransmogCache = true;
        for slot=1,150 do
            local id = oGetInventoryItemID("player", slot);
            if id then
                local cache = self:GetItemTransmogCache("player", slot);
                cache.ID = id;
                cache.Loaded = true;
            end
        end
        for bag=0,4 do
            for slot=1,36 do
                local id = GetContainerItemID(bag, slot);
                if id then
                    local cache = self:GetItemTransmogCache("player", bag, slot);
                    cache.ID = id;
                    cache.Loaded = true;
                end
            end
        end
    end,
    setEmptyBankTransmogCache = false,
    SetEmptyBankTransmogCache = function(self)
        if self.setEmptyBankTransmogCache then return; end
        self.setEmptyBankTransmogCache = true;
        for bag=5,11 do
            for slot=1,36 do
                local id = GetContainerItemID(bag, slot);
                if id then
                    local cache = self:GetItemTransmogCache("player", bag, slot);
                    if cache.ID == nil and not cache.Loaded then
                        cache.ID = id;
                        cache.Loaded = true;
                    end
                end
            end
        end
    end,
    lastInspectTarget = "",
    inspectFrameHooked = false,
    missingInspectItems = nil,
    dataRequestTooltip = CreateFrame("GameTooltip", ADDON_NAME.."DataRequestTooltip", UIParent),
    dataReadTooltip = CreateFrame("GameTooltip", ADDON_NAME.."DataReadTooltip", UIParent, "GameTooltipTemplate"),
};

-- --------------------------------------
-- Helper functions to manage collections
-- --------------------------------------
local function LoadList(container)
    return function(ids)
        for id in ids:gmatch("(%d+):") do
            container[tonumber(id)] = true;
        end
        if ends_with(ids, "END") then
            container.Loaded = true;
            if callback then
                callback(container);
            end
        end
    end;
end
local function LoadAllList(container, allContainer, callback)
    return function(ids)
        for id in ids:gmatch("(%d+):") do
            table.insert(container, tonumber(id))
            allContainer[tonumber(id)] = true;
        end
        if ends_with(ids, "END") then
            container.Loaded = true;
            allContainer.Loaded = true;
            for _, db in pairs(ezCollections.Cache.Slot) do
                if type(db) == "table" and not db.Loaded then
                    allContainer.Loaded = false;
                    break;
                end
            end
            if callback then
                callback(container, allContainer);
            end
        end
    end;
end
local function AddList(container, callback)
    return function(id)
        container[tonumber(id)] = true;
        if callback then
            callback(tonumber(id));
        end
    end;
end
local function RemoveList(container, callback)
    return function(id)
        container[tonumber(id)] = nil;
        if callback then
            callback(tonumber(id));
        end
    end;
end
local function ReloadList(request, container, callback)
    return function(ids)
        for id in pairs(container) do
            if type(id) == "number" then
                container[id] = nil;
            end
        end
        container.Loaded = false;
        if callback then
            callback(container);
        end
        ezCollections:SendAddonMessage(request);
    end;
end
local function ReloadAllList(request, container, allContainer, callback)
    return function(ids)
        for index, id in pairs(container) do
            if type(index) == "number" then
                allContainer[id] = nil;
                container[index] = nil;
            end
        end
        container.Loaded = false;
        allContainer.Loaded = false;
        if callback then
            callback(container, allContainer);
        end
        ezCollections:SendAddonMessage(request);
    end;
end
local function LoadItemTransmog(unit, slotStrings)
    for slotString in slotStrings:gmatch("(.-):") do
        local slots, itemString = strsplit("=", slotString, 2);
        local bag, slot = strsplit(",", slots, 2);
        local id, fakeEntry, fakeEnchant = strsplit(",", itemString, 3);
        bag = tonumber(bag);
        if slot ~= nil then
            slot = tonumber(slot) + 1;
        else
            bag = bag + 1;
        end
        local cache = ezCollections:GetItemTransmogCache(unit, bag, slot);
        cache.ID = tonumber(id);
        cache.FakeEntry = tonumber(fakeEntry);
        cache.FakeEnchant = fakeEnchant;
        cache.Loaded = true;
        cache.Loading = false;
    end
end

-- --------------------
-- Addon event handling
-- --------------------
function IsInspectFrameShown()
    return InspectFrame and InspectFrame:IsShown()
        or Examiner and Examiner:IsShown();
end
function addon:InitVersion()
    if self.versionRequestAttempts > 0 then
        self.versionRequestAttempts = self.versionRequestAttempts - 1;
        ezCollections:SendAddonMessage("VERSION:"..ADDON_VERSION);
    else
        self:CancelTimer(self.versionTimer);
        self.versionTimer = nil;
    end
end
function addon:UpdateInspect(unit)
    if not ezCollections.Config.RestoreItemIcons.Inspect then return; end

    -- GearScoreList requests inspects by hovering over players, which can screw up with us,
    -- since client can only hold one inspected unit in memory and we're expecting to update inspected slots later down the line
    if ezCollections.lastInspectRequestUnit ~= unit then
        ezCollections.lastInspectTarget = "";
        NotifyInspect("target");
        return;
    end

    ezCollections.missingInspectItems = { };
    if ezCollections.lastInspectTarget ~= "" and unit == ezCollections.lastInspectTarget and IsInspectFrameShown() then
        local InspectPaperDollItemSlotButton_Update = InspectPaperDollItemSlotButton_Update or function(...) end;
        InspectPaperDollItemSlotButton_Update(InspectHeadSlot);
        InspectPaperDollItemSlotButton_Update(InspectNeckSlot);
        InspectPaperDollItemSlotButton_Update(InspectShoulderSlot);
        InspectPaperDollItemSlotButton_Update(InspectBackSlot);
        InspectPaperDollItemSlotButton_Update(InspectChestSlot);
        InspectPaperDollItemSlotButton_Update(InspectShirtSlot);
        InspectPaperDollItemSlotButton_Update(InspectTabardSlot);
        InspectPaperDollItemSlotButton_Update(InspectWristSlot);
        InspectPaperDollItemSlotButton_Update(InspectHandsSlot);
        InspectPaperDollItemSlotButton_Update(InspectWaistSlot);
        InspectPaperDollItemSlotButton_Update(InspectLegsSlot);
        InspectPaperDollItemSlotButton_Update(InspectFeetSlot);
        InspectPaperDollItemSlotButton_Update(InspectFinger0Slot);
        InspectPaperDollItemSlotButton_Update(InspectFinger1Slot);
        InspectPaperDollItemSlotButton_Update(InspectTrinket0Slot);
        InspectPaperDollItemSlotButton_Update(InspectTrinket1Slot);
        InspectPaperDollItemSlotButton_Update(InspectMainHandSlot);
        InspectPaperDollItemSlotButton_Update(InspectSecondaryHandSlot);
        InspectPaperDollItemSlotButton_Update(InspectRangedSlot);
        local elvui = LibStub("AceAddon-3.0"):GetAddon("ElvUI", true);
        if elvui then
            local module = elvui:GetModule("Enhanced_PaperDoll");
            if module then
                module:UpdatePaperDoll("target");
            end
        end
        if oGlow and oGlow.updateInspect then
            oGlow.updateInspect();
        end
        if Examiner then
            local module = Examiner:GetModuleFromToken("ItemSlots")
            if module then
                module:UpdateItemSlots();
            end
        end
    end
    if ezCollections.missingInspectItems and #ezCollections.missingInspectItems > 0 then
        for i, id in pairs(ezCollections.missingInspectItems) do
            ezCollections.dataRequestTooltip:SetHyperlink("item:"..id);
            ezCollections.dataRequestTooltip:Hide();
        end
        self:ScheduleTimer("UpdateInspect", 1, unit);
    end
    ezCollections.missingInspectItems = nil;
end
function addon:PLAYER_LOGIN(event)
    self.versionRequestAttempts = 3;
    self:InitVersion();
    if self.versionTimer then
        self:CancelTimer(self.versionTimer);
    end
    self.versionTimer = self:ScheduleRepeatingTimer("InitVersion", 10);
end
function addon:CHAT_MSG_ADDON(event, prefix, message, distribution, sender)
    if prefix ~= ADDON_PREFIX or sender ~= "" then return; end

    match(message, "VERSIONCHECK", function(version)
        self:PLAYER_LOGIN(event);
    end);
    match(message, "SERVERVERSION:", function(version)
        self.versionRequestAttempts = 0;
        local version, result, url = strsplit(":", version, 3);
        if version == "DISABLED" then
            result = "DISABLED";
        end
        if result ~= "OK" then
            ezCollections.NewVersion = { Version = version, URL = url };
            if result == "DISABLED" then
                ezCollections.NewVersion.Disabled = true;
            elseif result ~= "COMPATIBLE" then
                ezCollections.NewVersion.Outdated = true;
            end
            if result ~= "DISABLED" or not ezCollections.Config.NewVersion.HideRetiredPopup then
                StaticPopup_Show("EZCOLLECTIONS_NEW_VERSION");
            end
        end
        if not ezCollections.Allowed and (result == "OK" or result == "COMPATIBLE") then
            ezCollections.Allowed = true;
            ezCollections:SendAddonMessage("GETTRANSMOG:ALL");
        end
    end);
    match(message, "CACHEVERSION:", function(version)
        if ezCollections.Cache.Version ~= tonumber(version) then
            ezCollections:ClearCache();
            ezCollections.Cache.Version = tonumber(version);
            for slot, db in pairs(ezCollections.Cache.Slot) do
                ezCollections:SendAddonMessage("LIST:ALL:"..slot);
            end
        end
    end);
    match(message, "UNLOCKSKINHINTCOMMAND:", function(command)
        ezCollections.UnlockSkinHintCommand = command;
    end);
    match(message, "COLLECTIONS:", function(collections)
        for k, collection in pairs({ strsplit(":", collections) }) do
            if collection ~= "END" then
                ezCollections:SendAddonMessage("LIST:"..collection);
            end
                if collection == "OWNEDITEM"        then ezCollections.Collections.OwnedItems.Enabled = true;
            elseif collection == "SKIN"             then ezCollections.Collections.Skins.Enabled = true;
            elseif collection == "TAKENQUEST"       then ezCollections.Collections.TakenQuests.Enabled = true;
            elseif collection == "REWARDEDQUEST"    then ezCollections.Collections.RewardedQuests.Enabled = true;
            end
        end
    end);
    match(message, "LIST:", function(list)
        match(list, "OWNEDITEM:",           LoadList(ezCollections.Collections.OwnedItems));
        match(list, "SKIN:",                LoadList(ezCollections.Collections.Skins));
        match(list, "TAKENQUEST:",          LoadList(ezCollections.Collections.TakenQuests));
        match(list, "REWARDEDQUEST:",       LoadList(ezCollections.Collections.RewardedQuests));
        for slot, db in pairs(ezCollections.Cache.Slot) do
            match(list, "ALL:"..slot..":",  LoadAllList(db, ezCollections.Cache.All));
        end
    end);
    match(message, "ADD:", function(list)
        match(list, "OWNEDITEM:",           AddList(ezCollections.Collections.OwnedItems));
        match(list, "SKIN:",                AddList(ezCollections.Collections.Skins, ezCollections.Alerts.AddSkin));
        match(list, "TAKENQUEST:",          AddList(ezCollections.Collections.TakenQuests));
        match(list, "REWARDEDQUEST:",       AddList(ezCollections.Collections.RewardedQuests));
    end);
    match(message, "REMOVE:", function(list)
        match(list, "OWNEDITEM:",           RemoveList(ezCollections.Collections.OwnedItems));
        match(list, "SKIN:",                RemoveList(ezCollections.Collections.Skins));
        match(list, "TAKENQUEST:",          RemoveList(ezCollections.Collections.TakenQuests));
        match(list, "REWARDEDQUEST:",       RemoveList(ezCollections.Collections.RewardedQuests));
    end);
    match(message, "RELOAD:", function(list)
        match(list, "OWNEDITEM:",           ReloadList("LIST:OWNEDITEM",     ezCollections.Collections.OwnedItems));
        match(list, "SKIN:",                ReloadList("LIST:SKIN",          ezCollections.Collections.Skins));
        match(list, "TAKENQUEST:",          ReloadList("LIST:TAKENQUEST",    ezCollections.Collections.TakenQuests));
        match(list, "REWARDEDQUEST:",       ReloadList("LIST:REWARDEDQUEST", ezCollections.Collections.RewardedQuests));
        for slot, db in pairs(ezCollections.Cache.Slot) do
            match(list, "ALL:"..slot..":",  ReloadAllList("LIST:ALL:"..slot, db, ezCollections.Cache.All));
        end
    end);
    match(message, "GETTRANSMOG:", function(data)
        if not match(data, "PLAYER:", function(nameSlotStrings)
            local unit, slotStrings = strsplit(":", nameSlotStrings, 2);
            ezCollections:ClearItemTransmogCache(unit);
            LoadItemTransmog(unit, slotStrings);
            ezCollections.AceAddon:UpdateInspect(unit);
        end) and not match(data, "ALL:", function(slotStrings)
            ezCollections:SetEmptyItemTransmogCache();
            LoadItemTransmog("player", slotStrings);
        end) then
            LoadItemTransmog("player", data);
        end
    end);
end
function addon:INSPECT_TALENT_READY(event)
    local target = GetUnitName("target");
    if not ezCollections.Allowed or not ezCollections.Config.RestoreItemIcons.Inspect or not target or not IsInspectFrameShown() then return; end
    local needs = self.needsInspectUpdate;
    self.needsInspectUpdate = nil;
    if target == ezCollections.lastInspectTarget then
        if needs then
            -- Must be delayed because for some unknown reason GetInventoryItem* will return old values here
            if not self.inspectUpdateTimer then
                self.inspectUpdateTimer = self:ScheduleTimer(function() self.inspectUpdateTimer = nil; self:UpdateInspect(target); end, 0.1);
            end
        end
        return;
    end

    if not ezCollections.inspectFrameHooked then
        if InspectFrame then
            InspectFrame:HookScript("OnHide", function(self)
                ezCollections.lastInspectTarget = "";
            end);
        end
        if Examiner then
            Examiner:HookScript("OnHide", function(self)
                ezCollections.lastInspectTarget = "";
            end);
        end
        ezCollections.inspectFrameHooked = true;
    end
    ezCollections.lastInspectTarget = target;
    ezCollections:ClearItemTransmogCache(target);
    ezCollections:SendAddonMessage("GETTRANSMOG:PLAYER:"..target);
end
function addon:UNIT_INVENTORY_CHANGED(event, unit)
    local target = GetUnitName(unit);
    if not ezCollections.Allowed or not ezCollections.Config.RestoreItemIcons.Inspect or not target or not IsInspectFrameShown() or target ~= ezCollections.lastInspectTarget then return; end

    if not self.reinspectTimer then
        self.reinspectTimer = self:ScheduleTimer(function() self.reinspectTimer = nil; NotifyInspect("target"); end, 0.5);
    end

    self.needsInspectUpdate = true;
    ezCollections:ClearItemTransmogCache(target);
    ezCollections:SendAddonMessage("GETTRANSMOG:PLAYER:"..target);
end
function addon:BANKFRAME_OPENED(event)
    if not ezCollections.Allowed then return; end

    ezCollections:SetEmptyBankTransmogCache();
end

-- ---------------------------------------------------------------------------
-- Replace transmogrified icons on paper doll frames with their original icons
-- ---------------------------------------------------------------------------
hooksecurefunc("NotifyInspect", function(unit)
    ezCollections.lastInspectRequestUnit = UnitName(unit);
end);
GetInventoryItemID = function(unit, slot, ...)
    if ezCollections.Allowed and ezCollections.Config.RestoreItemIcons.Equipment and UnitIsUnit(unit, "player") then
        -- Do nothing, GetInventoryItemID should be able to return real values for current player
    elseif ezCollections.Allowed and ezCollections.Config.RestoreItemIcons.Inspect and UnitIsUnit(unit, "target") and GetUnitName("target") == ezCollections.lastInspectTarget then
        local id = ezCollections:GetItemTransmogCache(unit, slot).ID;
        if id and id ~= 0 then
            return id;
        end
    end
    return oGetInventoryItemID(unit, slot, ...);
end
--[[ Called from secure code
local oGetInventoryItemLink = GetInventoryItemLink;
GetInventoryItemLink = function(unit, slot, ...)
    if ezCollections.Allowed and ezCollections.Config.RestoreItemIcons.Equipment and UnitIsUnit(unit, "player") then
        -- Do nothing, GetInventoryItemLink should be able to return real values for current player
    elseif ezCollections.Allowed and ezCollections.Config.RestoreItemIcons.Inspect and UnitIsUnit(unit, "target") and GetUnitName("target") == ezCollections.lastInspectTarget then
        local id = ezCollections:GetItemTransmogCache(unit, slot).ID;
        if id and id ~= 0 then
            local link = oGetInventoryItemLink(unit, slot, ...);
            if not link then
                local _, link = GetItemInfo(id);
                return link;
            end
            local parts = { strsplit(":", link) };
            parts[2] = id;
            return table.concat(parts, ":");
        end
    end
    return oGetInventoryItemLink(unit, slot, ...);
end
]]
local oGetInventoryItemTexture = GetInventoryItemTexture;
GetInventoryItemTexture = function(unit, slot, ...)
    if ezCollections.Allowed and ezCollections.Config.RestoreItemIcons.Equipment and UnitIsUnit(unit, "player") then
        local id = GetInventoryItemID(unit, slot);
        if not id or id == 0 then return; end
        local _, _, _, _, _, _, _, _, _, texture = GetItemInfo(id);
        return texture or oGetInventoryItemTexture(unit, slot, ...);
    elseif ezCollections.Allowed and ezCollections.Config.RestoreItemIcons.Inspect and UnitIsUnit(unit, "target") and GetUnitName("target") == ezCollections.lastInspectTarget then
        local id = ezCollections:GetItemTransmogCache(unit, slot).ID or GetInventoryItemID(unit, slot);
        if not id or id == 0 then return; end
        local _, _, _, _, _, _, _, _, _, texture = GetItemInfo(id);
        if not texture and ezCollections.missingInspectItems then
            table.insert(ezCollections.missingInspectItems, id);
        end
        return texture or oGetInventoryItemTexture(unit, slot, ...);
    end
    return oGetInventoryItemTexture(unit, slot, ...);
end

-- -------------------------------------------------------------------------------------
-- Inform the server that we're reloading UI and the addon might be disabled from now on
-- -------------------------------------------------------------------------------------
local oReloadUI = ReloadUI;
ReloadUI = function(...)
    if ezCollections.Allowed then
        ezCollections:SendAddonMessage("RELOADUI");
    end
    oReloadUI(...);
end

-- ------------------------------------------------------------
-- Add extra lines to tooltips about collection-related objects
-- ------------------------------------------------------------
local function GetTooltipItem(tooltip)
    local name, link = tooltip:GetItem();
    if not link then return; end
    local _, id, _ = strsplit(":", link);
    if not id then return; end
    id = tonumber(id);
    if id == ITEM_BACK or id == ITEM_HIDDEN then return; end
    return id, name, link;
end

local function TooltipHandlerItem(tooltip)
    if not ezCollections.Allowed then return; end

    local id = GetTooltipItem(tooltip);
    if not id then return; end

    local show = false;
    local color = ezCollections.Config.TooltipCollection.Color;
    local separator = not ezCollections.Config.TooltipCollection.Separator;
    if ezCollections.Config.TooltipCollection.OwnedItems and ezCollections:HasOwnedItem(id) == false then
        if not separator then separator = true; tooltip:AddLine(" "); end
        tooltip:AddLine(L["Tooltip.OwnedItems"], color.r, color.g, color.b, false);
        show = true;
    end
    if ezCollections.Config.TooltipCollection.Skins and ezCollections:IsSkinSource(id) == true and ezCollections:HasSkin(id) == false then
        if not separator then separator = true; tooltip:AddLine(" "); end
        tooltip:AddLine(L["Tooltip.Skins"], color.r, color.g, color.b, false);
        show = true;
    end
    if show then
        tooltip:Show();
    end
end
local function TooltipHandlerClear(tooltip)
    ezCollections.itemUnderCursor.ID = nil;
    ezCollections.itemUnderCursor.Bag = nil;
    ezCollections.itemUnderCursor.Slot = nil;
end
local function TooltipHandlerHyperlink(tooltip, link)
    if not ezCollections.Allowed then return; end

    local linkType, linkData = strsplit(":", link);
    if linkType == "quest" then
        local id = tonumber((linkData));

        local show = false;
        local color = ezCollections.Config.TooltipCollection.Color;
        local separator = not ezCollections.Config.TooltipCollection.Separator;
        if ezCollections.Config.TooltipCollection.TakenQuests and ezCollections:HasTakenQuest(id) == false then
            if not separator then separator = true; tooltip:AddLine(" "); end
            tooltip:AddLine(L["Tooltip.TakenQuests"], color.r, color.g, color.b, false);
            show = true;
        elseif ezCollections.Config.TooltipCollection.RewardedQuests and ezCollections:HasRewardedQuest(id) == false then
            if not separator then separator = true; tooltip:AddLine(" "); end
            tooltip:AddLine(L["Tooltip.RewardedQuests"], color.r, color.g, color.b, false);
            show = true;
        end
        if show then
            tooltip:Show();
        end
    end
end
local function TooltipHandlerInventory(tooltip, ...)
    if not ezCollections.Allowed then return; end

    local id = GetTooltipItem(tooltip);
    if not id then return; end

    if ezCollections:IsSkinSource(id) == true then
        local show = false;

        local unit, bag, slot;
        if type(select(1, ...)) == "string" then
            unit, bag = ...;
        else
            unit = "player";
            bag, slot = ...;
        end

        if ezCollections.Config.TooltipTransmog.Enable then
            local color = ezCollections.Config.TooltipTransmog.Color;
            local colorHex = "|cFF"..RGBPercToHex(color.r, color.g, color.b);
            local fakeEntry, fakeEnchant = ezCollections:GetItemTransmog(unit, bag, slot);
            if fakeEntry or fakeEnchant then
                local text = "";
                text = text..colorHex..L["Tooltip.Transmog.Header"].."|r";
                if fakeEntry and fakeEntry ~= 0 then
                    local name, _, _, _, _, _, _, _, _, texture = GetItemInfo(fakeEntry);
                    if fakeEntry == ITEM_HIDDEN then
                        name = L["Tooltip.Transmog.Entry.Hidden"];
                        texture = "Interface\\PaperDollInfoFrame\\UI-GearManager-LeaveItem-Transparent";
                    end
                    if not name or not texture then
                        name = "|cFFFF0000"..RETRIEVING_ITEM_INFO.."|r";
                        texture = "Interface\\Icons\\INV_Misc_QuestionMark";
                        ezCollections.dataRequestTooltip:SetHyperlink("item:"..fakeEntry);
                        ezCollections.dataRequestTooltip:Hide();
                    end
                    local size = ezCollections.Config.TooltipTransmog.IconEntry.Size or 0;
                    local crop = ezCollections.Config.TooltipTransmog.IconEntry.Crop and ":0:0:64:64:6:58:6:58" or "";
                    local icon = ezCollections.Config.TooltipTransmog.IconEntry.Enable and ("|T"..texture..":"..size..":"..size..crop.."|t ") or "";
                    text = text.."|n"..colorHex..icon..format(L["Tooltip.Transmog.EntryFormat"], name).."|r";
                end
                if fakeEnchant and fakeEnchant ~= "" then
                    local size = ezCollections.Config.TooltipTransmog.IconEnchant.Size or 0;
                    local icon = ezCollections.Config.TooltipTransmog.IconEnchant.Enable and "|TInterface\\AddOns\\ezCollections\\EnchantIcon:"..size..":"..size..":0:0:64:64:7:55:8:56|t " or "";
                    -- "|TInterface\\Icons\\INV_Scroll_05:0:0:0:0:64:64:6:58:6:58|t "
                    text = text.."|n"..colorHex..icon..format(L["Tooltip.Transmog.EnchantFormat"], fakeEnchant).."|r";
                end
                for i = 2,3 do
                    local line = _G[tooltip:GetName().."TextLeft"..i];
                    if line and line:GetText() and line:GetText():match(ITEM_LEVEL) then
                        line:SetText(line:GetText().."|n"..text);
                        line:SetNonSpaceWrap(false);
                        show = true;
                        break;
                    end
                end
                if not show then
                    for i = 2,5 do
                        local line = _G[tooltip:GetName().."TextLeft"..i];
                        if line and line:GetText() and not line:GetText():match(ITEM_HEROIC) and not line:GetText():match(ITEM_HEROIC_EPIC) then
                            line:SetText(text.."|n"..line:GetText());
                            line:SetNonSpaceWrap(false);
                            show = true;
                            break;
                        end
                    end
                end
            end
        end

        if ezCollections.Config.TooltipCollection.SkinUnlock and unit == "player" and ezCollections:HasSkin(id) == false then
            local color = ezCollections.Config.TooltipCollection.Color;
            local text;
            if GetBindingKey("EZCOLLECTIONS_UNLOCK_SKIN") then
                text = format(L["Tooltip.UnlockSkin.Binding"], table.concat({ GetBindingKey("EZCOLLECTIONS_UNLOCK_SKIN") }, L["Tooltip.UnlockSkin.Binding.Separator"]));
            else
                text = format(L["Tooltip.UnlockSkin.Command"], ezCollections.UnlockSkinHintCommand);
            end
            tooltip:AddLine(text, color.r * 0.75, color.g * 0.75, color.b * 0.75, false);
            ezCollections.itemUnderCursor.ID = id;
            ezCollections.itemUnderCursor.Bag = bag;
            ezCollections.itemUnderCursor.Slot = slot;
            show = true;
        end

        if ezCollections.Config.RestoreItemSets.Equipment and unit == "player" and bag >= 1 and bag <= 20 then
            for i = 1, 30 do
                local line = _G[tooltip:GetName().."TextLeft"..i];
                local line2 = _G[tooltip:GetName().."TextLeft"..(i+1)];
                local line3 = _G[tooltip:GetName().."TextLeft"..(i+2)];
                if line and line:GetText() and IsSameColor({ line:GetTextColor() }, 1, 0.8235, 0) and
                   line2 and line2:GetText() and (IsSameColor({ line2:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ line2:GetTextColor() }, 1, 1, 0.5922)) and
                   line3 and line3:GetText() and (IsSameColor({ line3:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ line3:GetTextColor() }, 1, 1, 0.5922)) then
                    local setName, count, max = line:GetText():match(FormatToPattern(ITEM_SET_NAME));
                    local _, link = GetItemInfo(id);
                    if setName and link then
                        local setHeader = nil;
                        local setItems = { };
                        local setBonuses = { };
                        local stooltip = ezCollections.dataReadTooltip;
                        stooltip:SetOwner(WorldFrame, "ANCHOR_NONE");
                        stooltip:ClearLines();
                        stooltip:SetHyperlink("item:"..id);
                        for j = 1, 30 do
                            local sline = _G[stooltip:GetName().."TextLeft"..j];
                            local sline2 = _G[stooltip:GetName().."TextLeft"..(j+1)];
                            local sline3 = _G[stooltip:GetName().."TextLeft"..(j+2)];
                            if sline and sline:GetText() and IsSameColor({ sline:GetTextColor() }, 1, 0.8235, 0) and
                               sline2 and sline2:GetText() and (IsSameColor({ sline2:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ sline2:GetTextColor() }, 1, 1, 0.5922)) and
                               sline3 and sline3:GetText() and (IsSameColor({ sline3:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ sline3:GetTextColor() }, 1, 1, 0.5922)) and
                               starts_with(sline:GetText(), setName) then
                                local bonuses = false;
                                for k = j, j + 30 do
                                    local sourceline = _G[stooltip:GetName().."TextLeft"..k];
                                    if sourceline and sourceline:GetText() then
                                        if not bonuses then
                                            if sourceline:GetText() == " " then
                                                bonuses = true;
                                            elseif IsSameColor({ sourceline:GetTextColor() }, 1, 0.8235, 0) then
                                                setHeader = sourceline:GetText();
                                            elseif IsSameColor({ sourceline:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ sourceline:GetTextColor() }, 1, 1, 0.5922) or IsSameColor({ sourceline:GetTextColor() }, 0, 1, 0) then
                                                setItems[sourceline:GetText()] = { sourceline:GetTextColor() };
                                            end
                                        elseif sourceline:GetText():match(FormatToPattern(ITEM_SET_BONUS)) or sourceline:GetText():match(FormatToPattern(ITEM_SET_BONUS_GRAY)) then
                                            if IsSameColor({ sourceline:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ sourceline:GetTextColor() }, 0, 1, 0) then
                                                table.insert(setBonuses, { sourceline:GetText(), { sourceline:GetTextColor() } });
                                            end
                                        else
                                            break;
                                        end
                                    end
                                end
                                break;
                            end
                        end
                        if setHeader then
                            show = true;
                            local bonuses = false;
                            for k = i, i + 30 do
                                local destline = _G[tooltip:GetName().."TextLeft"..k];
                                if destline and destline:GetText() then
                                    if not bonuses then
                                        if destline:GetText() == " " then
                                            bonuses = true;
                                        elseif IsSameColor({ destline:GetTextColor() }, 1, 0.8235, 0) then
                                            destline:SetText(setHeader);
                                        elseif IsSameColor({ destline:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ destline:GetTextColor() }, 1, 1, 0.5922) or IsSameColor({ destline:GetTextColor() }, 0, 1, 0) then
                                            local color = setItems[destline:GetText()];
                                            if color then
                                                destline:SetTextColor(unpack(color));
                                            end
                                        end
                                    elseif destline:GetText():match(FormatToPattern(ITEM_SET_BONUS)) or destline:GetText():match(FormatToPattern(ITEM_SET_BONUS_GRAY)) then
                                        if IsSameColor({ destline:GetTextColor() }, 0.5, 0.5, 0.5) or IsSameColor({ destline:GetTextColor() }, 0, 1, 0) then
                                            if #setBonuses > 0 then
                                                local bonus = table.remove(setBonuses, 1);
                                                destline:SetText(bonus[1]);
                                                destline:SetTextColor(unpack(bonus[2]));
                                            end
                                        end
                                    else
                                        break;
                                    end
                                end
                            end
                        end
                        stooltip:Hide();
                    end
                    break;
                end
            end
        end

        if show then
            tooltip:Show();
        end
    end
end

CreateFrame("Frame", ADDON_NAME.."TooltipHooker", UIParent):SetScript("OnUpdate", function(self, ...)
    local tooltips =
    {
        -- Game's own tooltips
        GameTooltip, ItemRefTooltip,
        -- Tooltips from other addons (feel free to add yours)
        AtlasLootTooltip, AtlasQuestTooltip, LightHeadedTooltip, MobMapTooltip, PWTooltip,
    };
    for k, tooltip in pairs(tooltips) do
        if tooltip then
            tooltip:HookScript("OnTooltipSetItem", TooltipHandlerItem);
            tooltip:HookScript("OnTooltipCleared", TooltipHandlerClear);
            hooksecurefunc(tooltip, "SetHyperlink", TooltipHandlerHyperlink);
            hooksecurefunc(tooltip, "SetInventoryItem", TooltipHandlerInventory);
            hooksecurefunc(tooltip, "SetBagItem", TooltipHandlerInventory);
        end
    end
    self:SetScript("OnUpdate", nil);
end);
